@@ -24,6 +24,8 @@
 
 #include "eti_source.h"
 #include "eti_player.h"
+#include "edi_source.h"
+#include "edi_player.h"
 #include "fic_decoder.h"
 #include "tools.h"
 #include "version.h"
@@ -32,7 +34,9 @@
 // --- DABlinTextOptions -----------------------------------------------------------------
 struct DABlinTextOptions {
 	std::string filename;
+	std::string source_format;
 	std::string initial_label;
+	bool initial_first_found_service;
 	int initial_sid;
 	int initial_scids;
 	int initial_subchid_dab;
@@ -46,6 +50,8 @@ struct DABlinTextOptions {
 	bool disable_dyn_fic_msgs;
 	int gain;
 DABlinTextOptions() :
+	source_format(EnsembleSource::FORMAT_ETI),
+	initial_first_found_service(false),
 	initial_sid(LISTED_SERVICE::sid_none),
 	initial_scids(LISTED_SERVICE::scids_none),
 	initial_subchid_dab(AUDIO_SERVICE::subchid_none),
@@ -61,18 +67,18 @@ DABlinTextOptions() :
 
 
 // --- DABlinText -----------------------------------------------------------------
-class DABlinText : ETISourceObserver, ETIPlayerObserver, FICDecoderObserver {
+class DABlinText : EnsembleSourceObserver, EnsemblePlayerObserver, FICDecoderObserver {
 private:
 	DABlinTextOptions options;
 
-	ETISource *eti_source;
-	ETIPlayer *eti_player;
+	EnsembleSource *ensemble_source;
+	EnsemblePlayer *ensemble_player;
 	FICDecoder *fic_decoder;
 
-	void ETIProcessFrame(const uint8_t *data) {eti_player->ProcessFrame(data);}
-	void ETIUpdateProgress(const ETI_PROGRESS& progress);
+	void EnsembleProcessFrame(const uint8_t *data) {ensemble_player->ProcessFrame(data);}
+	void EnsembleUpdateProgress(const ENSEMBLE_PROGRESS& progress);
 
-	void ETIProcessFIC(const uint8_t *data, size_t len) {fic_decoder->Process(data, len);}
+	void EnsembleProcessFIC(const uint8_t *data, size_t len) {fic_decoder->Process(data, len);}
 
 	void FICChangeService(const LISTED_SERVICE& service);
 	void FICDiscardedFIB();
@@ -80,8 +86,8 @@ public:
 	DABlinText(DABlinTextOptions options);
 	~DABlinText();
 
-	void DoExit() {eti_source->DoExit();}
-	int Main() {return eti_source->Main();}
+	void DoExit() {ensemble_source->DoExit();}
+	int Main() {return ensemble_source->Main();}
 };
 
 
