@@ -1,6 +1,6 @@
 /*
     DABlin - capital DAB experience
-    Copyright (C) 2015-2019 Stefan Pöschel
+    Copyright (C) 2015-2020 Stefan Pöschel
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -44,6 +44,7 @@ static void usage(const char* exe) {
 					"  -s <sid>     ID of the service to be played\n"
 					"  -x <scids>   ID of the service component to be played (requires service ID)\n"
 					"  -g <gain>    USB stick gain to pass to DAB live source (auto gain is default)\n"
+					"  -G           Use default gain for DAB live source (instead of auto gain)\n"
 					"  -r <path>    Path for recordings (default: %s)\n"
 					"  -P <size>    Recording prebuffer size in seconds (default: %zu)\n"
 					"  -p           Output PCM to stdout instead of using SDL\n"
@@ -51,7 +52,7 @@ static void usage(const char* exe) {
 					"  -I           Don't catch up on stream after interruption\n"
 					"  -S           Initially disable slideshow\n"
 					"  -L           Enable loose behaviour (e.g. PAD conformance)\n"
-					"  -F           Disable dynamic FIC messages (e.g. dynamic PTY)\n"
+					"  -F           Disable dynamic FIC messages (dynamic PTY, announcements)\n"
 					"  file         Input file to be played (stdin, if not specified)\n",
 					EnsembleSource::FORMAT_ETI.c_str(),
 					EnsembleSource::FORMAT_EDI.c_str(),
@@ -77,10 +78,11 @@ int main(int argc, char **argv) {
 
 	DABlinGTKOptions options;
 	int initial_param_count = 0;
+	int gain_param_count = 0;
 
 	// option args
 	int c;
-	while((c = getopt(argc, argv, "hf:d:D:C:c:l:g:r:P:s:x:1puISLF")) != -1) {
+	while((c = getopt(argc, argv, "hf:d:D:C:c:l:g:Gr:P:s:x:1puISLF")) != -1) {
 		switch(c) {
 		case 'h':
 			usage(argv[0]);
@@ -117,6 +119,11 @@ int main(int argc, char **argv) {
 			break;
 		case 'g':
 			options.gain = strtol(optarg, nullptr, 0);
+			gain_param_count++;
+			break;
+		case 'G':
+			options.gain = DAB_LIVE_SOURCE_CHANNEL::default_gain;
+			gain_param_count++;
 			break;
 		case 'r':
 			options.recordings_path = optarg;
@@ -204,11 +211,15 @@ int main(int argc, char **argv) {
 	}
 
 
-	// at most one initial param needed!
+	// at most one param needed!
 	if(initial_param_count > 1) {
 		fprintf(stderr, "At most one initial parameter shall be specified!\n");
 		usage(argv[0]);
 	}
+	if(gain_param_count > 1) {
+		fprintf(stderr, "At most one gain parameter shall be specified!\n");
+		usage(argv[0]);
+	}
 
 
 	fprint_dablin_banner(stderr);
@@ -238,6 +249,8 @@ DABlinGTK::DABlinGTK(DABlinGTKOptions options) {
 	rec_duration_ms = 0;
 	rec_prebuffer_filled_ms = 0;
 
+	dt_lto = FIC_ENSEMBLE::lto_none;
+
 	slideshow_window.set_transient_for(*this);
 
 	ensemble_update_progress.GetDispatcher().connect(sigc::mem_fun(*this, &DABlinGTK::EnsembleUpdateProgressEmitted));
@@ -247,6 +260,7 @@ DABlinGTK::DABlinGTK(DABlinGTKOptions options) {
 	pad_change_dynamic_label.GetDispatcher().connect(sigc::mem_fun(*this, &DABlinGTK::PADChangeDynamicLabelEmitted));
 	pad_change_slide.GetDispatcher().connect(sigc::mem_fun(*this, &DABlinGTK::PADChangeSlideEmitted));
 	do_rec_status_update.GetDispatcher().connect(sigc::mem_fun(*this, &DABlinGTK::DoRecStatusUpdateEmitted));
+	do_datetime_update.GetDispatcher().connect(sigc::mem_fun(*this, &DABlinGTK::DoDateTimeUpdateEmitted));
 
 	if(options.source_format == EnsembleSource::FORMAT_ETI)
 		ensemble_player = new ETIPlayer(options.pcm_output, options.untouched_output, options.disable_int_catch_up, this);
@@ -410,6 +424,19 @@ void DABlinGTK::InitWidgets() {
 	label_dl.set_valign(Gtk::ALIGN_START);
 	label_dl.set_padding(WIDGET_SPACE, WIDGET_SPACE);
 
+	frame_label_datetime.set_label("Local date/time");
+	frame_label_datetime.set_sensitive(false);
+	frame_label_datetime.add(label_datetime);
+	label_datetime.set_halign(Gtk::ALIGN_START);
+	label_datetime.set_padding(WIDGET_SPACE, WIDGET_SPACE);
+
+	frame_label_asu.set_label("Announcement support");
+	frame_label_asu.set_sensitive(false);
+	frame_label_asu.set_hexpand(true);
+	frame_label_asu.add(label_asu);
+	label_asu.set_halign(Gtk::ALIGN_START);
+	label_asu.set_padding(WIDGET_SPACE, WIDGET_SPACE);
+
 	progress_position.set_show_text();
 
 
@@ -430,7 +457,9 @@ void DABlinGTK::InitWidgets() {
 	top_grid.attach_next_to(tglbtn_mute, tglbtn_slideshow, Gtk::POS_RIGHT, 1, 1);
 	top_grid.attach_next_to(vlmbtn, tglbtn_mute, Gtk::POS_RIGHT, 1, 1);
 	top_grid.attach_next_to(frame_label_dl, frame_combo_channels, Gtk::POS_BOTTOM, 7, 1);
-	top_grid.attach_next_to(progress_position, frame_label_dl, Gtk::POS_BOTTOM, 7, 1);
+	top_grid.attach_next_to(frame_label_datetime, frame_label_dl, Gtk::POS_BOTTOM, 2, 1);
+	top_grid.attach_next_to(frame_label_asu, frame_label_datetime, Gtk::POS_RIGHT, 5, 1);
+	top_grid.attach_next_to(progress_position, frame_label_datetime, Gtk::POS_BOTTOM, 7, 1);
 
 	show_all_children();
 	progress_position.hide();	// invisible until progress updated
@@ -483,7 +512,7 @@ void DABlinGTK::SetService(const LISTED_SERVICE& service) {
 		std::string label = FICDecoder::ConvertLabelToUTF8(service.label, &charset_name);
 
 		set_title(label + " - DABlin");
-		set_icon_name("media-playback-start");
+		set_icon_name(tglbtn_record.get_active() ? "media-record" : "media-playback-start");
 
 		std::string tooltip_text =
 				"Short label: \"" + FICDecoder::DeriveShortLabelUTF8(label, service.label.short_label_mask) + "\"\n"
@@ -527,8 +556,11 @@ void DABlinGTK::SetService(const LISTED_SERVICE& service) {
 		tglbtn_record.set_sensitive(false);
 	}
 
+	UpdateAnnouncementSupport(service);
+
 	// if the audio service changed, reset format/DL/slide + switch
-	if(!ensemble_player->IsSameAudioService(service.audio_service)) {
+	bool audio_service_changed = !ensemble_player->IsSameAudioService(service.audio_service);
+	if(audio_service_changed) {
 		// stop playback of old service
 		if(options.rec_prebuffer_size_s > 0)
 			ensemble_player->RemoveUntouchedStreamConsumer(this);
@@ -565,7 +597,8 @@ void DABlinGTK::SetService(const LISTED_SERVICE& service) {
 			ensemble_player->AddUntouchedStreamConsumer(this);
 	}
 
-	if(service.HasSLS()) {
+	// (re)init SLS on new service (having SLS) only once
+	if(service.HasSLS() && (audio_service_changed || slideshow_window.IsEmptySlide())) {
 		slideshow_window.AwaitSlide();
 		if(tglbtn_slideshow.get_active())
 			slideshow_window.TryToShow();
@@ -573,6 +606,69 @@ void DABlinGTK::SetService(const LISTED_SERVICE& service) {
 	}
 }
 
+void DABlinGTK::UpdateAnnouncementSupport(const LISTED_SERVICE& service) {
+	// reset, if no service/announcement support
+	if(service.IsNone() || service.asu_flags == LISTED_SERVICE::asu_flags_none) {
+		frame_label_asu.set_sensitive(false);
+		frame_label_asu.set_tooltip_markup("");
+		label_asu.set_label("");
+		return;
+	}
+
+	// assemble text - iterate through all supported announcement types
+	std::string ac_str;
+	for(int type = 0; type < 16; type++) {
+		uint16_t type_flag = 1 << type;
+		if(!(service.asu_flags & type_flag))
+			continue;
+
+		// check all matching announcement clusters for active announcement
+		std::string asw_color;
+		for(const cids_t::value_type& cid : service.cids) {
+			asw_clusters_t::const_iterator it = ensemble.asw_clusters.find(cid);
+			if(it == ensemble.asw_clusters.cend() || !(it->second.asw_flags & type_flag))
+				continue;
+
+			if(it->second.subchid == service.audio_service.subchid) {
+				asw_color = "yellow";
+			} else {
+				if(asw_color.empty())
+					asw_color = "cyan";
+			}
+		}
+
+		if(!ac_str.empty())
+			ac_str += " + ";
+		if(!asw_color.empty())
+			ac_str += "<span bgcolor='" + asw_color + "'>";
+		ac_str += FICDecoder::ConvertASuTypeToString(type);
+		if(!asw_color.empty())
+			ac_str += "</span>";
+	}
+
+	// assemble tooltip - iterate through all matching announcement clusters
+	std::string ac_tooltip;
+	char cid_string[5];
+	for(const cids_t::value_type& cid : service.cids) {
+		asw_clusters_t::const_iterator it = ensemble.asw_clusters.find(cid);
+		bool present = it != ensemble.asw_clusters.cend();
+		bool active = present && (service.asu_flags & it->second.asw_flags);
+
+		ac_tooltip += ac_tooltip.empty() ? "Cluster(s): " : " / ";
+		if(active)
+			ac_tooltip += "<u>";
+		snprintf(cid_string, sizeof(cid_string), "0x%02X", cid);
+		ac_tooltip += std::string(cid_string);
+		if(present)
+			ac_tooltip += " (SubChId " + std::to_string(it->second.subchid) + ")";
+		if(active)
+			ac_tooltip += "</u>";
+	}
+
+	frame_label_asu.set_sensitive(true);
+	frame_label_asu.set_tooltip_markup(ac_tooltip);
+	label_asu.set_markup(ac_str);
+}
 
 void DABlinGTK::on_tglbtn_record() {
 	if(tglbtn_record.get_active()) {
@@ -932,6 +1028,17 @@ void DABlinGTK::EnsembleChangeFormatEmitted() {
 	label_format.set_label(ensemble_change_format.Pop().GetSummary());
 }
 
+void DABlinGTK::FICChangeEnsemble(const FIC_ENSEMBLE& ensemble) {
+	fic_change_ensemble.PushAndEmit(ensemble);
+
+	// update LTO for date/time + show
+	if(dt_lto != ensemble.lto) {
+		dt_lto = ensemble.lto;
+//		fprintf(stderr, "### LTO updated\n");
+		ShowDateTime(false);
+	}
+}
+
 void DABlinGTK::FICChangeEnsembleEmitted() {
 //	fprintf(stderr, "### FICChangeEnsembleEmitted\n");
 
@@ -946,6 +1053,8 @@ void DABlinGTK::FICChangeEnsembleEmitted() {
 			"EId: " + StringTools::IntToHex(ensemble.eid, 4);
 	if(ensemble.ecc != FIC_ENSEMBLE::ecc_none)
 		tooltip_text += "\n" "ECC: " + StringTools::IntToHex(ensemble.ecc, 2);
+	if(ensemble.lto != FIC_ENSEMBLE::lto_none)
+		tooltip_text += "\n" "LTO: " + FICDecoder::ConvertLTOToString(ensemble.lto);
 	if(ensemble.inter_table_id != FIC_ENSEMBLE::inter_table_id_none)
 		tooltip_text += "\n" "International table ID: " + StringTools::IntToHex(ensemble.inter_table_id, 2) + " (" + FICDecoder::ConvertInterTableIDToString(ensemble.inter_table_id) + ")";
 
@@ -999,6 +1108,59 @@ void DABlinGTK::FICChangeServiceEmitted() {
 	}
 }
 
+void DABlinGTK::FICChangeUTCDateTime(const FIC_DAB_DT& utc_dt) {
+	// (re)sync to DAB date/time + update/show/reschedule
+	utc_dt_next = utc_dt;
+	dt_update = std::chrono::steady_clock::now();
+//	fprintf(stderr, "### time synced\n");
+	ShowDateTime(true);
+}
+
+void DABlinGTK::ShowDateTime(bool scheduled) {
+	// if scheduled, update
+	if(scheduled)
+		utc_dt_curr = utc_dt_next;
+
+	// if UTC date/time and LTO set, show, if output changed
+	if(!utc_dt_curr.IsNone() && dt_lto != FIC_ENSEMBLE::lto_none) {
+		std::string dt_str = FICDecoder::ConvertDateTimeToString(utc_dt_curr, dt_lto, false);
+//		fprintf(stderr, "### time: %s\n", FICDecoder::ConvertDateTimeToString(utc_dt_curr, dt_lto, true).c_str());
+		if(dt_str_prev != dt_str) {
+			dt_str_prev = dt_str;
+//			fprintf(stderr, "### ----- time displayed: %s\n", dt_str.c_str());
+			DoDateTimeUpdate(dt_str);
+		}
+	}
+
+	// if scheduled, (re)schedule
+	if(scheduled) {
+		if(!utc_dt_next.IsMsNone()) {
+			// long form
+			dt_update += std::chrono::milliseconds(1000 - utc_dt_next.ms);
+			utc_dt_next.dt.tm_sec++;
+			utc_dt_next.ms = 0;
+		} else {
+			// short form
+			dt_update += std::chrono::minutes(1);
+			utc_dt_next.dt.tm_min++;
+		}
+		if(mktime(&utc_dt_next.dt) == (time_t) -1)
+			throw std::runtime_error("DABlinGTK: error while normalizing date/time");
+	}
+}
+
+void DABlinGTK::EnsembleDoRegularWork() {
+	// if UTC date/time set and scheduled time reached, update/show/reschedule
+	if(!utc_dt_curr.IsNone() && std::chrono::steady_clock::now() >= dt_update)
+		ShowDateTime(true);
+}
+
+void DABlinGTK::DoDateTimeUpdateEmitted() {
+	// display received date/time
+	frame_label_datetime.set_sensitive(true);
+	label_datetime.set_label(do_datetime_update.Pop());
+}
+
 void DABlinGTK::FICDiscardedFIB() {
 	fprintf(stderr, "\x1B[33m" "(FIB)" "\x1B[0m" " ");
 }
@@ -1020,6 +1182,13 @@ void DABlinGTK::on_combo_channels() {
 	label_ensemble.set_label("");
 	frame_label_ensemble.set_tooltip_text("");
 
+	frame_label_datetime.set_sensitive(false);
+	label_datetime.set_label("");
+
+	utc_dt_curr = FIC_DAB_DT();
+	dt_lto = FIC_ENSEMBLE::lto_none;
+	dt_str_prev = "";
+
 	// apply
 	Gtk::TreeModel::iterator row_it = combo_channels.get_active();
 	if(combo_channels_liststore->iter_is_valid(row_it)) {
