@@ -1,6 +1,6 @@
 /*
     DABlin - capital DAB experience
-    Copyright (C) 2015-2019 Stefan Pöschel
+    Copyright (C) 2015-2020 Stefan Pöschel
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -182,6 +182,7 @@ private:
 	FICDecoder *fic_decoder;
 	PADDecoder *pad_decoder;
 
+	// recording
 	std::mutex rec_mutex;
 	FILE* rec_file;
 	std::string rec_filename;
@@ -189,6 +190,13 @@ private:
 	rec_samples_t rec_prebuffer;
 	long int rec_prebuffer_filled_ms;
 
+	// date/time
+	FIC_DAB_DT utc_dt_curr;
+	int dt_lto;
+	std::string dt_str_prev;
+	FIC_DAB_DT utc_dt_next;
+	std::chrono::steady_clock::time_point dt_update;
+
 	Gtk::TreeModel::iterator resume_channel_it;
 	int resume_service_sid;
 	int resume_service_scids;
@@ -198,6 +206,7 @@ private:
 	void EnsembleProcessFrame(const uint8_t *data) {ensemble_player->ProcessFrame(data);}
 	void EnsembleUpdateProgress(const ENSEMBLE_PROGRESS& progress) {ensemble_update_progress.PushAndEmit(progress);}
 	void EnsembleUpdateProgressEmitted();
+	void EnsembleDoRegularWork();
 
 	// ensemble data change
 	GTKDispatcherQueue<AUDIO_SERVICE_FORMAT> ensemble_change_format;
@@ -244,6 +253,12 @@ private:
 	Gtk::Frame frame_label_dl;
 	Gtk::Label label_dl;
 
+	Gtk::Frame frame_label_datetime;
+	Gtk::Label label_datetime;
+
+	Gtk::Frame frame_label_asu;
+	Gtk::Label label_asu;
+
 	Gtk::ProgressBar progress_position;
 
 
@@ -252,6 +267,8 @@ private:
 	void AddChannel(const dab_channels_t::value_type& dab_channel, int gain);
 
 	void SetService(const LISTED_SERVICE& service);
+	void UpdateAnnouncementSupport(const LISTED_SERVICE& service);
+	void ShowDateTime(bool scheduled);
 
 	void on_btn_channels_stop();
 	void on_tglbtn_record();
@@ -272,15 +289,20 @@ private:
 	void DoRecStatusUpdateEmitted();
 	void UpdateRecStatus(bool decoding);
 
+	GTKDispatcherQueue<std::string> do_datetime_update;
+	void DoDateTimeUpdate(const std::string& dt_str) {do_datetime_update.PushAndEmit(dt_str);}
+	void DoDateTimeUpdateEmitted();
+
 	// FIC data change
 	GTKDispatcherQueue<FIC_ENSEMBLE> fic_change_ensemble;
-	void FICChangeEnsemble(const FIC_ENSEMBLE& ensemble) {fic_change_ensemble.PushAndEmit(ensemble);}
+	void FICChangeEnsemble(const FIC_ENSEMBLE& ensemble);
 	void FICChangeEnsembleEmitted();
 
 	GTKDispatcherQueue<LISTED_SERVICE> fic_change_service;
 	void FICChangeService(const LISTED_SERVICE& service) {fic_change_service.PushAndEmit(service);}
 	void FICChangeServiceEmitted();
 
+	void FICChangeUTCDateTime(const FIC_DAB_DT& utc_dt);
 	void FICDiscardedFIB();
 
 	// PAD data change
