@@ -1,6 +1,6 @@
 /*
     DABlin - capital DAB experience
-    Copyright (C) 2015-2019 Stefan Pöschel
+    Copyright (C) 2015-2020 Stefan Pöschel
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -24,6 +24,7 @@ void FICDecoder::Reset() {
 	ensemble = FIC_ENSEMBLE();
 	services.clear();
 	subchannels.clear();
+	utc_dt = FIC_DAB_DT();
 }
 
 void FICDecoder::Process(const uint8_t *data, size_t len) {
@@ -101,12 +102,21 @@ void FICDecoder::ProcessFIG0(const uint8_t *data, size_t len) {
 	case 9:
 		ProcessFIG0_9(data, len);
 		break;
+	case 10:
+		ProcessFIG0_10(data, len);
+		break;
 	case 13:
 		ProcessFIG0_13(data, len);
 		break;
 	case 17:
 		ProcessFIG0_17(data, len);
 		break;
+	case 18:
+		ProcessFIG0_18(data, len);
+		break;
+	case 19:
+		ProcessFIG0_19(data, len);
+		break;
 //	default:
 //		fprintf(stderr, "FICDecoder: received unsupported FIG 0/%d with %zu field bytes\n", header.extension, len);
 	}
@@ -305,20 +315,21 @@ void FICDecoder::ProcessFIG0_8(const uint8_t *data, size_t len) {
 
 void FICDecoder::ProcessFIG0_9(const uint8_t *data, size_t len) {
 	// FIG 0/9 - Time and country identifier - Country, LTO and International table
-	// ensemble ECC and international table ID only
+	// ensemble ECC/LTO and international table ID only
 
 	if(len < 3)
 		return;
 
 	FIC_ENSEMBLE new_ensemble = ensemble;
+	new_ensemble.lto = (data[0] & 0x20 ? -1 : 1) * (data[0] & 0x1F);
 	new_ensemble.ecc = data[1];
 	new_ensemble.inter_table_id = data[2];
 
 	if(ensemble != new_ensemble) {
 		ensemble = new_ensemble;
 
-		fprintf(stderr, "FICDecoder: ECC: 0x%02X, international table ID: 0x%02X (%s)\n",
-				ensemble.ecc, ensemble.inter_table_id, ConvertInterTableIDToString(ensemble.inter_table_id).c_str());
+		fprintf(stderr, "FICDecoder: ECC: 0x%02X, LTO: %s, international table ID: 0x%02X (%s)\n",
+				ensemble.ecc, ConvertLTOToString(ensemble.lto).c_str(), ensemble.inter_table_id, ConvertInterTableIDToString(ensemble.inter_table_id).c_str());
 
 		UpdateEnsemble();
 
@@ -331,6 +342,56 @@ void FICDecoder::ProcessFIG0_9(const uint8_t *data, size_t len) {
 	}
 }
 
+void FICDecoder::ProcessFIG0_10(const uint8_t *data, size_t len) {
+	// FIG 0/10 - Date and time (d&t)
+
+	if(len < 4)
+		return;
+
+	FIC_DAB_DT new_utc_dt;
+
+	// retrieve date
+	int mjd = (data[0] & 0x7F) << 10 | data[1] << 2 | data[2] >> 6;
+
+	int y0 = floor((mjd - 15078.2) / 365.25);
+	int m0 = floor((mjd - 14956.1 - floor(y0 * 365.25)) / 30.6001);
+	int d = mjd - 14956 - floor(y0 * 365.25) - floor(m0 * 30.6001);
+	int k = (m0 == 14 || m0 == 15) ? 1 : 0;
+	int y = y0 + k;
+	int m = m0 - 1 - k * 12;
+
+	new_utc_dt.dt.tm_year = y;		// from 1900
+	new_utc_dt.dt.tm_mon = m - 1;	// 0-based
+	new_utc_dt.dt.tm_mday = d;
+
+	// retrieve time
+	bool utc_flag = data[2] & 0x08;
+	new_utc_dt.dt.tm_hour = (data[2] & 0x07) << 2 | data[3] >> 6;
+	new_utc_dt.dt.tm_min = data[3] & 0x3F;
+	new_utc_dt.dt.tm_isdst = -1;	// ignore DST
+	if(utc_flag) {
+		// long form
+		if(len < 6)
+			return;
+		new_utc_dt.dt.tm_sec = data[4] >> 2;
+		new_utc_dt.ms = (data[4] & 0x03) << 8 | data[5];
+	} else {
+		// short form
+		new_utc_dt.dt.tm_sec = 0;
+		new_utc_dt.ms = FIC_DAB_DT::ms_none;
+	}
+
+	if(utc_dt != new_utc_dt) {
+		// print only once
+		if(utc_dt.IsNone())
+			fprintf(stderr, "FICDecoder: UTC date/time: %s\n", ConvertDateTimeToString(new_utc_dt, 0, true).c_str());
+
+		utc_dt = new_utc_dt;
+
+		observer->FICChangeUTCDateTime(utc_dt);
+	}
+}
+
 void FICDecoder::ProcessFIG0_13(const uint8_t *data, size_t len) {
 	// FIG 0/13 - User application information
 	// programme services only
@@ -414,6 +475,80 @@ void FICDecoder::ProcessFIG0_17(const uint8_t *data, size_t len) {
 	}
 }
 
+void FICDecoder::ProcessFIG0_18(const uint8_t *data, size_t len) {
+	// FIG 0/18 - Announcement support
+
+	// iterate through all services
+	for(size_t offset = 0; offset < len;) {
+		uint16_t sid = data[offset] << 8 | data[offset + 1];
+		uint16_t asu_flags = data[offset + 2] << 8 | data[offset + 3];
+		size_t number_of_clusters = data[offset + 4] & 0x1F;
+		offset += 5;
+
+		cids_t cids;
+		for(size_t i = 0; i < number_of_clusters; i++)
+			cids.emplace(data[offset++]);
+
+		FIC_SERVICE& service = GetService(sid);
+		uint16_t& current_asu_flags = service.asu_flags;
+		cids_t& current_cids = service.cids;
+		if(current_asu_flags != asu_flags || current_cids != cids) {
+			current_asu_flags = asu_flags;
+			current_cids = cids;
+
+			std::string cids_str;
+			char cid_string[5];
+			for(const cids_t::value_type& cid : cids) {
+				if(!cids_str.empty())
+					cids_str += "/";
+				snprintf(cid_string, sizeof(cid_string), "0x%02X", cid);
+				cids_str += std::string(cid_string);
+			}
+
+			fprintf(stderr, "FICDecoder: SId 0x%04X: ASu flags 0x%04X, cluster(s) %s\n",
+					sid, asu_flags, cids_str.c_str());
+
+			UpdateService(service);
+		}
+	}
+}
+
+void FICDecoder::ProcessFIG0_19(const uint8_t *data, size_t len) {
+	// FIG 0/19 - Announcement switching
+
+	// iterate through all announcement clusters
+	for(size_t offset = 0; offset < len;) {
+		uint8_t cid = data[offset];
+		uint16_t asw_flags = data[offset + 1] << 8 | data[offset + 2];
+		bool region_flag = data[offset + 3] & 0x40;
+		int subchid = data[offset + 3] & 0x3F;
+		offset += region_flag ? 5 : 4;
+
+		FIC_ASW_CLUSTER ac;
+		ac.asw_flags = asw_flags;
+		ac.subchid = subchid;
+
+		FIC_ASW_CLUSTER& current_ac = ensemble.asw_clusters[cid];
+		if(current_ac != ac) {
+			current_ac = ac;
+
+			if(!disable_dyn_msgs) {
+				fprintf(stderr, "FICDecoder: ASw cluster 0x%02X: flags 0x%04X, SubChId %2d\n",
+						cid, asw_flags, subchid);
+			}
+
+			UpdateEnsemble();
+
+			// update services that changes may affect
+			for(const fic_services_t::value_type& service : services) {
+				const FIC_SERVICE& s = service.second;
+				if(s.cids.find(cid) != s.cids.cend())
+					UpdateService(s);
+			}
+		}
+	}
+}
+
 void FICDecoder::ProcessFIG1(const uint8_t *data, size_t len) {
 	if(len < 1) {
 		fprintf(stderr, "FICDecoder: received empty FIG 1\n");
@@ -573,6 +708,8 @@ void FICDecoder::UpdateListedService(const FIC_SERVICE& service, int scids, bool
 	ls.label = service.label;
 	ls.pty_static = service.pty_static;
 	ls.pty_dynamic = service.pty_dynamic;
+	ls.asu_flags = service.asu_flags;
+	ls.cids = service.cids;
 	ls.pri_comp_subchid = service.pri_comp_subchid;
 	ls.multi_comps = multi_comps;
 
@@ -715,6 +852,12 @@ const char* FICDecoder::ptys_rbds_0x00_to_0x1D[] = {
 		"(rfu)", "Weather"
 };
 
+const char* FICDecoder::asu_types_0_to_10[] = {
+		"Alarm", "Road Traffic flash", "Transport flash", "Warning/Service",
+		"News flash", "Area weather flash", "Event announcement", "Special event",
+		"Programme Information", "Sport report", "Financial report"
+};
+
 std::string FICDecoder::ConvertLanguageToString(const int value) {
 	if(value >= 0x00 && value <= 0x2B)
 		return languages_0x00_to_0x2B[value];
@@ -725,6 +868,12 @@ std::string FICDecoder::ConvertLanguageToString(const int value) {
 	return "unknown (" + std::to_string(value) + ")";
 }
 
+std::string FICDecoder::ConvertLTOToString(const int value) {
+	char lto_string[7];
+	snprintf(lto_string, sizeof(lto_string), "%+03d:%02d", value / 2, (value % 2) ? 30 : 0);
+	return lto_string;
+}
+
 std::string FICDecoder::ConvertInterTableIDToString(const int value) {
 	switch(value) {
 	case 0x01:
@@ -736,6 +885,40 @@ std::string FICDecoder::ConvertInterTableIDToString(const int value) {
 	}
 }
 
+std::string FICDecoder::ConvertDateTimeToString(FIC_DAB_DT utc_dt, const int lto, bool output_ms) {
+	const char* weekdays[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
+
+	// if desired, apply LTO
+	if(lto)
+		utc_dt.dt.tm_min += lto * 30;
+
+	// normalize time (apply LTO, set day of week)
+	if(mktime(&utc_dt.dt) == (time_t) -1)
+		throw std::runtime_error("FICDecoder: error while normalizing date/time");
+
+	std::string result;
+	char s[11];
+
+	strftime(s, sizeof(s), "%F", &utc_dt.dt);
+	result += std::string(s) + ", " + weekdays[utc_dt.dt.tm_wday] + " - ";
+
+	if(!utc_dt.IsMsNone()) {
+		// long form
+		strftime(s, sizeof(s), "%T", &utc_dt.dt);
+		result += s;
+		if(output_ms) {
+			snprintf(s, sizeof(s), ".%03d", utc_dt.ms);
+			result += s;
+		}
+	} else {
+		// short form
+		strftime(s, sizeof(s), "%R", &utc_dt.dt);
+		result += s;
+	}
+
+	return result;
+}
+
 std::string FICDecoder::ConvertPTYToString(const int value, const int inter_table_id) {
 	switch(inter_table_id) {
 	case 0x01:
@@ -747,6 +930,12 @@ std::string FICDecoder::ConvertPTYToString(const int value, const int inter_tabl
 	}
 }
 
+std::string FICDecoder::ConvertASuTypeToString(const int value) {
+	if(value >= 0 && value <= 10)
+		return asu_types_0_to_10[value];
+	return "unknown (" + std::to_string(value) + ")";
+}
+
 std::string FICDecoder::DeriveShortLabelUTF8(const std::string& long_label, uint16_t short_label_mask) {
 	std::string short_label;
 
